/*
 * sbuf.c -- simple text editor buffer routines
 *
 * Part of Publib.  See manpage for more information.
 * "@(#)publib-sbuf:sbuf.c,v 1.2 1994/10/06 11:15:27 liw Exp"
 */


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "publib/sbuf.h"
#include "publib/errormsg.h"


/* Return length of gap */
#define gaplen(buf)	(buf->alloc - buf->len)



/* Create a buffer. */
Sbuf *sbuf_create(void) {
	Sbuf *buf;

	buf = malloc(sizeof(Sbuf));
	if (buf == NULL) {
		__publib_error("malloc failed");
		return NULL;
	}

	buf->block = NULL;
	buf->locked = 0;
	buf->gappos = 0;
	buf->alloc = 0;
	buf->len = 0;
	buf->marks = NULL;
	buf->markalloc = 0;

	sbuf_validate(buf);

	return buf;
}



/* Destroy a buffer */
void sbuf_destroy(Sbuf *buf) {
	sbuf_validate(buf);

	free(buf->block);
	free(buf->marks);
	free(buf);
}



/* Check that all the data structures for the buffer are OK.  Abort program
   if they aren't.  */
void (sbuf_validate)(Sbuf *buf) {
	struct __sbufmark *m;
	size_t i;

	assert(buf != NULL);
	assert(buf->len <= buf->alloc);
	assert(buf->block != NULL || buf->alloc == 0);
	assert(buf->gappos <= buf->alloc);
	assert(buf->len <= buf->alloc);

	assert(buf->marks != NULL || buf->markalloc == 0);
	for (i = 0; i < buf->markalloc; ++i) {
		m = buf->marks + i;
		if (m->inuse) {
			if (m->begin + m->len > buf->len) {
				printf("buf=%p, m=%u, m->begin=%ld, m->len=%ld, buf->len=%u\n",
					buf, i, m->begin, m->len, buf->len);
				fflush(stdout);
			}
			assert(m->begin + m->len <= buf->len);
		}
	}
}



/* Move gap before the pos'th character in text, and make sure it is at
   least len characters. */
int sbuf_movegap(Sbuf *buf, size_t pos, size_t len) {
	char *p;

	sbuf_validate(buf);
	assert(pos <= buf->len);
	assert(!buf->locked);

	if (gaplen(buf) < len || buf->block == NULL) {
		if (buf->block != NULL && sbuf_movegap(buf, buf->len, 0) < 0) {
			__publib_error("recursive sbuf_failed (can't happen)");
			return -1;
		}
		p = realloc(buf->block, buf->alloc + len + 4);
		if (p == NULL) {
			__publib_error("realloc failed");
			return -1;
		}
		buf->block = p;
		buf->alloc += len;
	}

	assert(buf->block != NULL);
        if (pos < buf->gappos)
               	memmove(buf->block + pos + gaplen(buf),
                       	buf->block + pos,
                       	buf->gappos - pos);
        else
               	memmove(buf->block + buf->gappos,
                       	buf->block + buf->gappos + gaplen(buf),
                       	pos - buf->gappos);
        buf->gappos = pos;

	sbuf_validate(buf);

	return 0;
}



/* Lock buffer against changes and return pointer to a contiguous block
   of memory holding the buffer contents (which may be modified except
   its length may not be modified).  */
char *sbuf_lock(Sbuf *buf) {
	sbuf_validate(buf);
	assert(!buf->locked);

	if (sbuf_movegap(buf, buf->len, 0) == -1) {
		__publib_error("sbuf_movegap failed");
		return NULL;
	}
	buf->locked = 1;
	return buf->block;
}



/* Release a lock on a buffer.  The user may not assume that the pointer
   returned by sbuf_lock is usable any longer.  */
void sbuf_unlock(Sbuf *buf) {
	sbuf_validate(buf);
	assert(buf->locked);

	buf->locked = 0;
}



/* Return length (== number of character in the buffer). */
long sbuf_length(Sbuf *buf) {
	sbuf_validate(buf);
	return buf->len;
}



/* Return the character at position pos.  Return EOF if just after last
   character.  */
int (sbuf_charat)(Sbuf *buf, long pos) {
	sbuf_validate(buf);
	assert(pos >= 0);
	assert(pos <= buf->len);
	assert(!buf->locked);

	if (pos == buf->len)
		return EOF;

	assert(buf->block != NULL);
	if (pos < buf->gappos)
		return buf->block[pos];
	return buf->block[pos + gaplen(buf)];
}



/* Create a mark. */
Sbufmark *sbuf_mark(Sbuf *buf, long pos, long len) {
	struct __sbufmark *p;
	Sbufmark *q;
	size_t i;

	sbuf_validate(buf);
	assert(pos >= 0);
	assert(len >= 0);
	assert(pos + len <= buf->len);

	p = NULL;
	for (i = 0; i < buf->markalloc; ++i) {
		if (!buf->marks[i].inuse) {
			p = buf->marks + i;
			break;
		}
	}
	if (p == NULL) {
		i = buf->markalloc + 8;
		p = realloc(buf->marks, i * sizeof(struct __sbufmark));
		if (p == NULL) {
			__publib_error("realloc failed");
			return NULL;
		}
		buf->marks = p;
		p = buf->marks + buf->markalloc;
		while (buf->markalloc < i)
			buf->marks[buf->markalloc++].inuse = 0;
		buf->markalloc = i;
	}

	q = malloc(sizeof(Sbufmark));
	if (q == NULL) {
		__publib_error("malloc failed");
		return NULL;
	}

	p->inuse = 1;
	p->begin = pos;
	p->len = len;
	p->dirty = 0;

	q->buf = buf;
	q->mark = p - buf->marks;

	sbuf_validate(buf);
	sbuf_validate_mark(q);

	return q;
}



/* Destroy a mark.  */
void sbuf_unmark(Sbufmark *mark) {
	sbuf_validate_mark(mark);

	mark->buf->marks[mark->mark].inuse = 0;
}



/* Check that a mark's data structures are ok.  */
void (sbuf_validate_mark)(Sbufmark *mark) {
	assert(mark != NULL);
	sbuf_validate(mark->buf);
	assert(mark->mark < mark->buf->markalloc);
	assert(mark->buf->marks[mark->mark].inuse);
}



/* Move an existing mark. */
void sbuf_remark(Sbufmark *mark, long pos, long len) {
	sbuf_validate_mark(mark);
	assert(pos >= 0);
	assert(len >= 0);
	assert(pos + len <= mark->buf->len);

	mark->buf->marks[mark->mark].begin = pos;
	mark->buf->marks[mark->mark].len = len;
	mark->buf->marks[mark->mark].dirty = 0;
}



/* Return starting position of mark. */
long sbuf_mark_begin(Sbufmark *mark) {
	sbuf_validate_mark(mark);
	return mark->buf->marks[mark->mark].begin;
}



/* Return end position of mark.  */
long sbuf_mark_end(Sbufmark *mark) {
	sbuf_validate_mark(mark);
	return mark->buf->marks[mark->mark].begin +
               mark->buf->marks[mark->mark].len;
}



/* Return length of mark.  */
long sbuf_mark_len(Sbufmark *mark) {
	sbuf_validate_mark(mark);
	return mark->buf->marks[mark->mark].len;
}



/* Return dirty bit.  */
int sbuf_mark_is_dirty(Sbufmark *mark) {
	sbuf_validate_mark(mark);
	return mark->buf->marks[mark->mark].dirty;
}



/* Set dirty bit.  */
void sbuf_mark_set_dirty(Sbufmark *mark, int dirty) {
	sbuf_validate_mark(mark);
	mark->buf->marks[mark->mark].dirty = dirty;
}



/* Copy marked text to a string.  */
void sbuf_strat(char *str, Sbufmark *mark) {
	struct __sbufmark *m;
	char *p;

	assert(str != NULL);
	sbuf_validate_mark(mark);

	p = sbuf_lock(mark->buf);
	m = mark->buf->marks + mark->mark;
	if (m->len > 0)
		memcpy(str, p + m->begin, m->len);
	str[m->len] = '\0';
	sbuf_unlock(mark->buf);
}



/* Replace the contents of a mark with a string.  */
int sbuf_strchange(Sbufmark *mark, const char *str, size_t len) {
	struct __sbufmark *m, old;
	long mend, oldend;
	size_t i;

	sbuf_validate_mark(mark);
	assert(str != NULL);
	assert(strlen(str) >= len);	/* bug if buffer may hold '\0'! */
	assert(!mark->buf->locked);

	m = mark->buf->marks + mark->mark;

	/*
	 * Delete old contents of mark
	 */
	if (sbuf_movegap(mark->buf, m->begin, len) == -1) {
		__publib_error("sbuf_movegap failed");
		return -1;
	}
	assert(m->len <= mark->buf->len - m->begin);
	mark->buf->len -= m->len;

	/*
	 * Insert string
	 */
	memcpy(mark->buf->block + mark->buf->gappos, str, len);
	mark->buf->len += len;
	mark->buf->gappos += len;

	/*
	 * Adjust marks
	 */
	old = *m;
	oldend = old.begin + old.len;
	for (i = 0; i < mark->buf->markalloc; ++i) {
		m = mark->buf->marks + i;
		if (!m->inuse)
			continue;
		mend = m->begin + m->len;
		if (m->begin < old.begin && mend > old.begin) {
			/* m starts before, extends into or after old */
			if (mend <= oldend)
				m->len = len + old.begin - m->begin;
			else
				m->len = m->len - old.len + len;
		} else if (m->begin >= old.begin && m->begin < oldend) {
			/* m starts within or at beginning of old */
			m->begin = old.begin;
			if (mend > oldend)
				m->len = len + (mend - oldend);
			else
				m->len = len;
		} else if (m->begin >= oldend) {
			/* m starts after or at end of old */
			m->len = m->len - old.len + len;
		}
		m->dirty = 1;
	}

	sbuf_validate_mark(mark);

	return 0;
}



/* Replace contents of one mark with those of another mark.  The marks may
   overlap.  */
int sbuf_change(Sbufmark *mark, Sbufmark *repl) {
	char *temp, tempbuf[10240];
	long len;
	int ret;

	sbuf_validate_mark(mark);
	sbuf_validate_mark(repl);

	len = sbuf_mark_len(repl);
	if (len < sizeof tempbuf)
		temp = tempbuf;
	else {
		temp = malloc(len+1);
		if (temp == NULL) {
			__publib_error("malloc failed");
			return -1;
		}
	}

	sbuf_strat(temp, repl);
	ret = sbuf_strchange(mark, temp, len);

	if (temp != tempbuf)
		free(temp);

	return ret;
}



/* Replace contents of mark m with the contents of file f */
int sbuf_insert_file(Sbufmark *m, FILE *f) {
	size_t i, n, inc;
	char *p, *q;

	n = 0;
	inc = 128*1024;	/* a nice small number */
	p = NULL;
	do {
		q = realloc(p, n + inc);
		if (q == NULL) {
			__publib_error("realloc failed");
			return -1;
		}
		p = q;

		i = fread(p+n, 1, inc, f);
		if (i != inc && ferror(f)) {
			__publib_error("fread failed");
			free(p);
			return -1;
		}

		n += i;
	} while (!feof(f));

	if (sbuf_strchange(m, p, n) == -1) {
		__publib_error("sbuf_strchange failed");
		free(p);
		return -1;
	}

	free(p);
	return 0;
}



/* Write contents of mark m to file f.  */
int sbuf_write_to_file(Sbufmark *m, FILE *f) {
	size_t i;
	long n;
	char *p;

	n = sbuf_mark_len(m);
	p = malloc(n);
	if (p == NULL) {
		__publib_error("malloc failed");
		return -1;
	}

	sbuf_strat(p, m);

	i = fwrite(p, 1, n, f);
	if (i != n) {
		__publib_error("fwrite failed");
		free(p);
		return -1;
	}

	free(p);
	return 0;
}

